package com.jpsoft.bus.modules.bus.service.impl;

import com.github.pagehelper.Page;
import com.jpsoft.bus.modules.bus.dto.StationStatusDTO;
import com.jpsoft.bus.modules.bus.entity.*;
import com.jpsoft.bus.modules.bus.service.*;
import com.jpsoft.bus.modules.common.utils.CommonUtil;
import com.jpsoft.gps.utils.GPSUtil;
import com.jpsoft.bus.config.GpsParamConfig;
import com.jpsoft.bus.modules.bus.dao.VehicleInfoDAO;
import com.jpsoft.bus.modules.common.dto.Sort;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.ibatis.annotations.Mapper;
import org.gavaghan.geodesy.Ellipsoid;
import org.gavaghan.geodesy.GlobalCoordinates;
import org.joda.time.DateTime;
import org.joda.time.Minutes;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.ValueOperations;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RestTemplate;

import javax.annotation.Resource;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.*;
import java.util.concurrent.TimeUnit;

@Mapper
@Component(value = "gpsService")
@Slf4j
public class GpsServiceImpl implements GpsService {
    @Resource(name = "vehicleInfoDAO")
    private VehicleInfoDAO vehicleInfoDAO;

    @Autowired
    private RestTemplate restTemplate;

    @Autowired
    private StationSubInfoService stationSubInfoService;

    @Autowired
    private VehicleInfoService vehicleInfoService;

    @Autowired
    private RouteInfoService routeInfoService;

    @Autowired
    private StationInfoService stationInfoService;

    @Autowired
    private ShiftInfoService shiftInfoService;

    @Autowired
    private GpsParamConfig gpsParamConfig;

    @Autowired
    private ValueOperations<String, Object> valueOperations;

    @Autowired
    private GpsDataInfoService gpsDataInfoService;

    /*
     * 批量实时更新车辆定位
     * @param vList 车辆实体列表
     * */
    @Override
    public List<Map<String, String>> queryLatestVehicleGps(List<VehicleInfo> vList) {
        // TODO Auto-generated method stub
        List<Map<String, String>> coordinateList = new ArrayList<Map<String, String>>();

        for (VehicleInfo v : vList) {
            Map<String, String> coordinate = new HashMap<String, String>();

            String gps = (String) valueOperations.get(v.getGpsDeviceNo() + "_gps");

            if (StringUtils.isNotEmpty(gps)) {
                String licenseNo = v.getLicensePlateNumber();

                String[] arr = gps.split(",");
                String longitude = arr[0];
                String latitude = arr[1];

                coordinate.put("jd", arr[0]);
                coordinate.put("wd", arr[1]);

                String address = "";

                if (Double.valueOf(longitude) > 0 && Double.valueOf(latitude) > 0) {
                    String key = "lbs_" + longitude + "," + latitude;

                    address = (String) valueOperations.get(key);

                    if (StringUtils.isEmpty(address)) {
                        address = GPSUtil.regeo(longitude, latitude, gpsParamConfig.getKey());
                        valueOperations.set(key, address, 1, TimeUnit.DAYS);
                    }

                    v.setLatestAddress(address);
                }

                coordinate.put("address", address);

                coordinate.put("licensePlateNumber", licenseNo);//车牌号
                coordinateList.add(coordinate);
            }
        }

        return coordinateList;
    }

    /*
     * 获取车辆轨迹
     * @param v 车辆实体
     * @param st 开始时间
     * @param et 结束时间
     * */
    @Override
    public List<Map<String, String>> getVehicleGpsList(VehicleInfo v, Date st, Date et) {
        // TODO Auto-generated method stub
        //开始时间不得大于结束时间，并且查询天数不得大于7天。
        List<Map<String, String>> gpsList = new ArrayList<Map<String, String>>();
        long day1 = st.getTime();
        long day2 = et.getTime();

        if (day1 > day2) {
            log.warn("查询车辆轨迹时开始时间大于结束时间");
            return null;
        }
        if (day2 - day1 >= 7 * 1000 * 3600 * 24) {
            log.warn("查询车辆轨迹时天数超过7天");
            return null;
        }

        Map<String, Object> searchParams = new HashMap<>();
        searchParams.put("deviceNo", v.getGpsDeviceNo());
        searchParams.put("startTime", st);
        searchParams.put("endTime", et);

        List<Sort> sortList = new ArrayList<>();
        sortList.add(new Sort("create_time", "asc"));

        List<GpsDataInfo> dataList = gpsDataInfoService.pageSearch(searchParams, 1, 1000, false, sortList);

        for (GpsDataInfo data : dataList) {
            Map<String, String> map = new HashMap<String, String>();

            map.put("lat", data.getLatitude());//纬度
            map.put("lng", data.getLongitude());//经度

            gpsList.add(map);
        }

        return gpsList;
    }

    /*
     * 判断车辆是否在线，true:在线；false:熄火
     * @param v 车辆实体
     * */
    @Override
    public boolean isVehicleOnline(VehicleInfo v) {
        //最后一次上传经纬度时间是否超过5分钟
        GpsDataInfo gpsDataInfo = gpsDataInfoService.findLast(v.getGpsDeviceNo());

        if (gpsDataInfo != null) {
            DateTime lastTime = new DateTime(gpsDataInfo.getCreateTime());

            long minutes = Minutes.minutesBetween(lastTime, DateTime.now()).getMinutes();

            return minutes < 5;
        } else {
            return false;
        }
    }

    @Override
    public List<Map<String, String>> queryVehicleOnlineInfo(List<VehicleInfo> vList) {
        // TODO Auto-generated method stub
        //单次查询,车牌号查询
        List<Map<String, String>> onlineList = new ArrayList<>();

        for (VehicleInfo v : vList) {
            Map<String, String> olmap = new HashMap<String, String>();

            boolean online = isVehicleOnline(v);

            if (online) {
                olmap.put("online", "1");
            } else {
                olmap.put("online", "0");
            }

            olmap.put("licensePlateNumber", v.getLicensePlateNumber());//车牌号
            onlineList.add(olmap);
        }

        return onlineList;
    }

    /*
     * 查询车辆ACC状态，0表示ACC关闭1表示ACC开启。
     * @param vList 车辆实体列表
     * */
    @Override
    public List<Map<String, String>> queryVehicleAccInfo(List<VehicleInfo> vList) {
        // TODO Auto-generated method stub
        List<Map<String, String>> accList = new ArrayList<Map<String, String>>();

        for (VehicleInfo v : vList) {
            Map<String, String> olmap = new HashMap<String, String>();

            boolean online = isVehicleOnline(v);

            if (online) {
                olmap.put("acc", "1");
            } else {
                olmap.put("acc", "0");
            }

            olmap.put("licensePlateNumber", v.getLicensePlateNumber());//车牌号
            accList.add(olmap);
        }

        return accList;
    }

    @Override
    public StationStatusDTO queryStation(String id) throws Exception {

        StationStatusDTO stationStatusDTO = new StationStatusDTO();
        VehicleInfo vehicleInfo = vehicleInfoService.get(id);
        if (vehicleInfo == null) {
            throw new Exception("车辆不存在");
        }
        List<ShiftInfo> shiftInfoList = shiftInfoService.findByVehicleIdAndStatus(id, "1");
        if (shiftInfoList.size() == 0) {
            throw new Exception("此车辆无正在运行的班次");
        }
        ShiftInfo shiftInfo = shiftInfoList.get(0);

        RouteInfo routeInfo = routeInfoService.get(shiftInfo.getRouteId());

        List<StationInfo> stationInfoList = stationInfoService.findByRouteId(routeInfo.getId());

        //线路中的当前站点
         StationInfo currentStation = stationInfoService.get(shiftInfo.getCurrentStationId());

         //下一站站点
         StationInfo nextStation = null;

         Integer index = stationInfoList.indexOf(currentStation);

         Integer nextIndex = index+1;

         //反向行驶
         if (!stationInfoList.get(0).getId().equals(shiftInfo.getStartStationId())){
             nextIndex = index-1;
         }

         if (nextIndex>=0 && nextIndex < stationInfoList.size()){
             nextStation = stationInfoList.get(nextIndex);
         }

         if(StringUtils.isNotEmpty(vehicleInfo.getLatitude())
         && StringUtils.isNotEmpty(vehicleInfo.getLatitude())
         && currentStation!=null) {
             //车辆离当前站点的距离
             GlobalCoordinates source = new GlobalCoordinates(Double.valueOf(vehicleInfo.getLatitude()), Double.valueOf(vehicleInfo.getLongitude()));
             GlobalCoordinates target = new GlobalCoordinates(Double.valueOf(currentStation.getLatitude()), Double.valueOf(currentStation.getLongitude()));

             double meter1 = CommonUtil.getDistanceMeter(source, target, Ellipsoid.Sphere);

             if (meter1 <= Double.valueOf(50)) {
                 stationStatusDTO.setNowStationStatus("1");
             } else {
                 stationStatusDTO.setNowStationStatus("2");
             }
         }

        stationStatusDTO.setNowStationId(currentStation.getId());
        stationStatusDTO.setNowStationName(currentStation.getName());

        if (nextStation!=null) {
            stationStatusDTO.setNextStationId(nextStation.getId());
            stationStatusDTO.setNextStationName(nextStation.getName());

            if (nextStation.equals(currentStation)){
                stationStatusDTO.setNextStationStatus(stationStatusDTO.getNowStationStatus());
            }else {
                stationStatusDTO.setNextStationStatus("3");
            }
        }

        return stationStatusDTO;
    }

    @Override
    public String getLatelyStation(String id,boolean manual) throws Exception {
        VehicleInfo vehicleInfo = vehicleInfoService.get(id);
        if (vehicleInfo == null) {
            throw new Exception("车辆不存在");
        }

        String currentStationId = null;
        List<ShiftInfo> shiftInfoList = shiftInfoService.findByVehicleIdAndStatus(id, "1");

        if (shiftInfoList.size() > 0) {
            ShiftInfo shiftInfo = shiftInfoList.get(0);

            RouteInfo routeInfo = routeInfoService.get(shiftInfo.getRouteId());

            List<StationInfo> stationInfoList = stationInfoService.findByRouteId(routeInfo.getId());
            currentStationId = shiftInfo.getCurrentStationId();

            int currentIndex = 0;

            if(StringUtils.isNotEmpty(currentStationId)){
                //当前站点
                StationInfo currentStation = stationInfoService.get(currentStationId);

                if(currentStation!=null) {
                    currentIndex = currentStation.getSortNo();
                }
            }

            //线路终点站
            StationInfo endStation = stationInfoService.get(shiftInfo.getEndStationId());

            boolean frontToEnd = true;
        //    int distance = 100;

            if (stationInfoList.size() > 0) {
                if(endStation.getId().equals(stationInfoList.get(0).getId())){
                    //反向行驶
                    frontToEnd = false;
                }

                if(frontToEnd){
                    int n = 0;
                    for (int i=0;i<stationInfoList.size();i++) {
                        StationInfo stationInfo = stationInfoList.get(i);

                        if(stationInfo.getSortNo() <= currentIndex){
                            //正向行驶时，排序比当前站小的站点不用计算
                            continue;
                        }

                        int distance = stationInfo.getRadius();

                        if (stationInfo.getRadius() == null){
                            distance = 100;
                        }

                        if(matchDistance(stationInfo,shiftInfo.getStartStationId(),shiftInfo.getEndStationId(),vehicleInfo,distance)){
                           currentStationId = stationInfo.getId();
                           break;
                        }

                        //非人为设置当前站点则只查后两站
                        if (!manual) {
                            n++;

                            if (n >= 2) {
                                break;
                            }
                        }
                    }
                }
                else{
                    int n = 0;
                    for (int i=stationInfoList.size() - 1;i>=0;i--) {
                        StationInfo stationInfo = stationInfoList.get(i);

                        if (stationInfo.getSortNo() >= currentIndex){
                            //反向行驶时，排序比当前站大的站点不用计算
                            continue;
                        }

                        int distance = stationInfo.getRadius();
                        if (stationInfo.getRadius() == null){
                            distance = 100;
                        }

                        if(matchDistance(stationInfo,shiftInfo.getStartStationId(),shiftInfo.getEndStationId(),vehicleInfo,distance)){
                            currentStationId = stationInfo.getId();
                            break;
                        }

                        //非人为设置当前站点则只查后两站
                        if (!manual) {
                            n++;
                            if (n >= 2) {
                                break;
                            }
                        }
                    }
                }

                //gps修改站点时，查询是否有附属站点
                /*if (!manual){

                    Double lo = Double.valueOf(vehicleInfo.getLongitude());
                    Double la = Double.valueOf(vehicleInfo.getLatestAddress());

                    //50米
                    Double maxLo = lo + 0.0005D;
                    Double minLo = lo - 0.0005D;
                    Double maxLa = la + 0.0004D;
                    Double minLa = la - 0.0004D;


                    Map<String, Object> searchParams = new HashMap<>();
                    searchParams.put("maxLongitude", maxLo);
                    searchParams.put("minLongitude", minLo);
                    searchParams.put("maxLatitude", maxLa);
                    searchParams.put("minLatitude", minLa);
                    searchParams.put("startStationId",shiftInfo.getStartStationId());
                    searchParams.put("endStationId",shiftInfo.getEndStationId());
                    List<Sort> sortList = new ArrayList<>();
                    sortList.add(new Sort("create_time", "desc"));
                    Page<StationSubInfo> page = stationSubInfoService.pageSearch(searchParams,1,1,false,sortList);
                    List<StationSubInfo> stationSubInfoList = page.getResult();
                    if (stationSubInfoList.size()>0){
                        StationSubInfo stationSubInfo = stationSubInfoList.get(0);
                        currentStationId = stationSubInfo.getStationId();
                    }
                }*/

            }
        }
        return currentStationId;
    }

    @Override
    public boolean matchDistance(StationInfo stationInfo,VehicleInfo vehicleInfo,int distance){
        GlobalCoordinates source = new GlobalCoordinates(Double.valueOf(vehicleInfo.getLatitude()), Double.valueOf(vehicleInfo.getLongitude()));
        GlobalCoordinates target = new GlobalCoordinates(Double.valueOf(stationInfo.getLatitude()), Double.valueOf(stationInfo.getLongitude()));

        double meter1 = CommonUtil.getDistanceMeter(source, target, Ellipsoid.Sphere);

        if (meter1 <= Double.valueOf(distance)) {
            return true;
        }
        else{
            return false;
        }
    }

    public boolean matchDistance(StationInfo stationInfo,String startStationId,String endStationId,VehicleInfo vehicleInfo,int distance){
        StationSubInfo stationSubInfo = stationSubInfoService.findByStationIdStartEnd(stationInfo.getId(),startStationId,endStationId);

        if (stationSubInfo != null){
            StationInfo subStationInfo = new StationInfo();
            subStationInfo.setLongitude(stationSubInfo.getLongitude());
            subStationInfo.setLatitude(stationSubInfo.getLatitude());

            if(matchDistance(subStationInfo,vehicleInfo,stationSubInfo.getRadius())){
                return true;
            }
        }

        return matchDistance(stationInfo,vehicleInfo,distance);
    }

    public String encode(String text) {
        try {
            text = URLEncoder.encode(text, "UTF-8");
        } catch (NumberFormatException | UnsupportedEncodingException e) {
            System.out.println(e);
        }

        return text;
    }

    public static void main(String[] args) {
        int s1 = -2147480189;

        System.out.println(Integer.toBinaryString(s1));

        int pos = 2;
        int value = 1 << (pos - 1);

        System.out.println(value);

        if ((s1 & value) == value) {
            System.out.println("ACC开启");
        }
    }
}